<?php

class WC_GZDP_VAT_Validation_Rest {

	private $api_url = 'https://ec.europa.eu/taxation_customs/vies/rest-api/';

	private $options = array(
		'debug'            => false,
		'requester_vat_id' => array(
			'country' => '',
			'number'  => '',
		),
	);

	private $valid = false;

	private $data = array();

	private $errors = false;

	public function __construct( $options = array() ) {
		$this->options = wp_parse_args( $options, $this->options );
	}

	public function check( $country, $nr ) {
		$this->errors = new WP_Error();
		$instance     = WC_germanized_pro();

		try {
			$url      = untrailingslashit( trailingslashit( $this->api_url ) . '/check-vat-number' );
			$response = wp_remote_post(
				esc_url_raw( $url ),
				array(
					'headers' => array( 'Content-Type' => 'application/json' ),
					'timeout' => 100,
					'body'    => wp_json_encode(
						array(
							'countryCode'              => $country,
							'vatNumber'                => $nr,
							'requesterMemberStateCode' => $this->options['requester_vat_id']['country'],
							'requesterNumber'          => $this->options['requester_vat_id']['number'],
						),
						JSON_PRETTY_PRINT
					),
				)
			);

			if ( false !== $response ) {
				if ( is_wp_error( $response ) ) {
					throw new Exception( $response->get_error_message() );
				}

				$code = wp_remote_retrieve_response_code( $response );
				$body = wp_remote_retrieve_body( $response );

				if ( 200 === $code ) {
					$json_body = json_decode( $body, true );

					if ( $json_body ) {
						if ( true === $json_body['valid'] ) {
							$this->valid = true;
							$this->data  = array(
								'name'       => $this->parse_string( isset( $json_body['name'] ) ? $json_body['name'] : '' ),
								'identifier' => $this->parse_string( isset( $json_body['requestIdentifier'] ) ? $json_body['requestIdentifier'] : '' ), // phpcs:ignore WordPress.NamingConventions.ValidVariableName.UsedPropertyNotSnakeCase
								'company'    => $this->parse_string( isset( $json_body['name'] ) ? $json_body['name'] : '' ), // phpcs:ignore WordPress.NamingConventions.ValidVariableName.UsedPropertyNotSnakeCase
								'address'    => $this->parse_string( isset( $json_body['address'] ) ? $json_body['address'] : '' ), // phpcs:ignore WordPress.NamingConventions.ValidVariableName.UsedPropertyNotSnakeCase
								'date'       => date_i18n( 'Y-m-d H:i:s' ),
								'raw'        => $json_body,
								'vat_id'     => $country . $nr,
							);
						} else {
							$instance->log( sprintf( 'VAT is invalid: %s', $country . $nr ), 'info', 'vat-validation' );

							$this->valid = false;
							$this->data  = array();

							$this->errors->add( 'vat-id-invalid', __( 'The VAT ID you\'ve provided is invalid.', 'woocommerce-germanized-pro' ) );
						}
					}
				} else {
					$json_body = json_decode( $body, true );
					$error     = 'Missing error message';
					$code      = 500;

					if ( $json_body && isset( $json_body['errorWrappers'] ) ) {
						$error = $json_body['errorWrappers']['message'];
						$code  = $json_body['errorWrappers']['error'];
					}

					throw new Exception( $error, $code );
				}
			}
		} catch ( Exception $e ) {
			$this->valid = false;
			$this->data  = array();

			$instance->log( sprintf( 'REST Error (%s) while performing VAT ID validation: %s', $e->getCode(), $e->getMessage() ), 'error', 'vat-validation' );
			$this->errors->add( 'vat-request-error', __( 'There was an error while validating your VAT ID. Please try again in a few minutes.', 'woocommerce-germanized-pro' ) );
		}

		return apply_filters( 'woocommerce_gzdp_vat_validation_rest_result', $this->valid, $country, $nr );
	}

	/**
	 * @return bool|WP_Error
	 */
	public function get_error_messages() {
		return ( is_wp_error( $this->errors ) && wc_gzd_wp_error_has_errors( $this->errors ) ) ? $this->errors : false;
	}

	public function is_valid() {
		return $this->valid;
	}

	public function get_name() {
		return isset( $this->data['name'] ) ? $this->data['name'] : '';
	}

	public function get_company() {
		return isset( $this->data['company'] ) ? $this->data['company'] : '';
	}

	public function get_identifier() {
		return isset( $this->data['identifier'] ) ? $this->data['identifier'] : '';
	}

	public function get_date() {
		return isset( $this->data['date'] ) ? $this->data['date'] : '';
	}

	public function get_formatted_address() {
		$address = array_filter( array( trim( $this->get_name() ), trim( $this->get_address() ) ) );

		return implode( ' ', $address );
	}

	public function get_raw() {
		return isset( $this->data['raw'] ) ? $this->data['raw'] : '';
	}

	public function get_data() {
		return $this->data;
	}

	public function get_address() {
		return isset( $this->data['address'] ) ? $this->data['address'] : '';
	}

	public function is_debug() {
		return ( true === $this->options['debug'] );
	}

	private function parse_string( $string ) {
		return ( '---' !== $string ? preg_replace( '/\s+/u', ' ', $string ) : false );
	}
}


