<?php

namespace Vendidero\StoreaBill\eInvoice;

use Vendidero\StoreaBill\BusinessInformation;
use Vendidero\StoreaBill\Interfaces\EInvoice;

defined( 'ABSPATH' ) || exit;

abstract class Invoice implements EInvoice {

	protected $profile = '';

	protected $buyer = null;

	protected $seller = null;

	protected $ship_to = null;

	protected $invoice_type = '';

	public function __construct( $profile, $invoice_type = 'invoice' ) {
		if ( ! in_array( $profile, $this->get_available_profiles(), true ) ) {
			throw new \Exception( sprintf( '%s is not a valid profile', $profile ), 500 );
		}

		$this->profile      = $profile;
		$this->invoice_type = $invoice_type;
		$this->buyer        = new Buyer();
		$this->seller       = new Seller();
		$this->ship_to      = new ShipTo();
	}

	public function get_profile() {
		return $this->profile;
	}

	public function supports_pdf_attachment() {
		return false;
	}

	public function get_pdf( $pdf, $xml = '' ) {
		return null;
	}

	public function get_available_profiles() {
		return Helper::get_profiles_by_specification( $this->get_specification() );
	}

	public function set_additional_fields( $fields ) {
		$available_fields = array();

		if ( $profile_data = Helper::get_profile( $this->get_specification(), $this->get_profile() ) ) {
			$available_fields = $profile_data['fields'];
		}

		foreach ( $available_fields as $field_name => $field ) {
			$field = wp_parse_args(
				$field,
				array(
					'mappings' => array(),
					'id'       => '',
				)
			);

			if ( isset( $fields[ $field['id'] ] ) ) {
				$field_value = $fields[ $field['id'] ];

				if ( ! empty( $field['mappings'] ) ) {
					foreach ( $field['mappings'] as $mapping ) {
						$mapping_exp = explode( ':', $mapping );
						$object      = $mapping_exp[0];
						$key         = $mapping_exp[1];

						if ( ! empty( $object ) && ! empty( $key ) ) {
							switch ( $object ) {
								case 'seller':
									$this->get_seller()->set_address( array( $key => $field_value ) );
									break;
								case 'buyer':
									$this->get_buyer()->set_address( array( $key => $field_value ) );
									$this->get_ship_to()->set_address( array( $key => $field_value ) );
									break;
								case 'ship_to':
									$this->get_ship_to()->set_address( array( $key => $field_value ) );
									break;
								case 'invoice':
									$setter            = 'set' === substr( $key, 0, 3 ) ? $key : "set_{$key}";
									$setter_camel_case = 'set' === substr( $key, 0, 3 ) ? $key : "set{$key}";

									if ( $this->is_callable( $setter ) ) {
										$this->{ $setter }( $field_value );
									} elseif ( $this->is_callable( $setter_camel_case ) ) {
										$this->{ $setter_camel_case }( $field_value );
									}

									break;
							}
						}
					}
				}
			}
		}
	}

	public function get_invoice_type() {
		return $this->invoice_type;
	}

	/**
	 * @param \Vendidero\StoreaBill\Invoice\Invoice $invoice
	 *
	 * @return void
	 */
	public function from_invoice( $invoice, $additional_fields = array() ) {
		$this->invoice_type = $invoice->get_invoice_type();
		$buyer_address      = array_merge(
			$invoice->get_address(),
			array(
				'customer_id' => $invoice->get_customer_id() ? $invoice->get_customer_id() : '',
			)
		);

		$this->buyer->set_address( $buyer_address );

		$ship_to_address = array_merge(
			$invoice->get_shipping_address(),
			array(
				'customer_id' => $invoice->get_customer_id() ? $invoice->get_customer_id() : '',
			)
		);

		$this->ship_to->set_address( $ship_to_address );

		$this->seller->set_address(
			array(
				'first_name'       => BusinessInformation::get_contact_first_name(),
				'last_name'        => BusinessInformation::get_contact_last_name(),
				'email'            => BusinessInformation::get_contact_email(),
				'company'          => BusinessInformation::get_base_company_name(),
				'address_1'        => BusinessInformation::get_base_address(),
				'address_2'        => BusinessInformation::get_base_address_2(),
				'postcode'         => BusinessInformation::get_base_postcode(),
				'city'             => BusinessInformation::get_base_city(),
				'country'          => BusinessInformation::get_base_country(),
				'state'            => BusinessInformation::get_base_state(),
				'vat_id'           => BusinessInformation::get_vat_id(),
				'local_tax_number' => BusinessInformation::get_local_tax_number(),
			)
		);

		$this->set_additional_fields( $additional_fields );
	}

	/**
	 * @param \Vendidero\StoreaBill\Invoice\Invoice $invoice
	 *
	 * @return string[]
	 */
	protected function get_notes_by_invoice( $invoice ) {
		$notes = array();

		if ( $invoice->get_order_number() ) {
			$notes[] = sprintf( _x( 'Order %s', 'storeabill-core', 'woocommerce-germanized-pro' ), $invoice->get_order_number() );
		}

		if ( 'cancellation' === $invoice->get_invoice_type() ) {
			$notes[] = sprintf( _x( 'Cancellation to %s', 'storeabill-core', 'woocommerce-germanized-pro' ), $invoice->get_parent_formatted_number() );

			if ( $invoice->has_refund_order() ) {
				$notes[] = sprintf( _x( 'Refund %s', 'storeabill-core', 'woocommerce-germanized-pro' ), $invoice->get_refund_order_number() );
			}
		}

		if ( $invoice->get_payment_method_title() ) {
			if ( $invoice->is_paid() ) {
				$notes[] = sprintf( _x( 'Paid via %s', 'storeabill-core', 'woocommerce-germanized-pro' ), $invoice->get_payment_method_title() );
			} else {
				$notes[] = sprintf( _x( 'Awaiting payment via %s', 'storeabill-core', 'woocommerce-germanized-pro' ), $invoice->get_payment_method_title() );
			}
		}

		if ( $invoice->get_payment_transaction_id() ) {
			$notes[] = sprintf( _x( 'Payment transaction id %s', 'storeabill-core', 'woocommerce-germanized-pro' ), $invoice->get_payment_transaction_id() );
		}

		return $notes;
	}

	/**
	 * @return Buyer
	 */
	public function get_buyer() {
		return $this->buyer;
	}

	/**
	 * @return Seller
	 */
	public function get_seller() {
		return $this->seller;
	}

	public function get_ship_to() {
		return $this->ship_to;
	}

	public function is_callable( $method ) {
		if ( is_callable( array( $this, $method ) ) ) {
			return true;
		}

		return false;
	}
}
