<?php

namespace Vendidero\StoreaBill\eInvoice\ZugFeRD;

use Vendidero\StoreaBill\BusinessInformation;
use Vendidero\StoreaBill\Invoice\Item;
use Vendidero\StoreaBill\Package;
use Vendidero\StoreaBill\Utilities\Numbers;
use Vendidero\StoreaBill\Vendor\horstoeko\zugferd\codelists\ZugferdPaymentMeans;
use Vendidero\StoreaBill\Vendor\horstoeko\zugferd\ZugferdDocumentBuilder;
use Vendidero\StoreaBill\Vendor\horstoeko\zugferd\ZugferdDocumentPdfMerger;
use Vendidero\StoreaBill\Vendor\horstoeko\zugferd\ZugferdDocumentPdfReader;
use Vendidero\StoreaBill\Vendor\horstoeko\zugferd\ZugferdDocumentReader;
use Vendidero\StoreaBill\Vendor\horstoeko\zugferd\ZugferdProfiles;
use Vendidero\StoreaBill\Vendor\horstoeko\zugferd\ZugferdSettings;
use Vendidero\StoreaBill\Vendor\horstoeko\zugferd\ZugferdXsdValidator;

defined( 'ABSPATH' ) || exit;

class Invoice extends \Vendidero\StoreaBill\eInvoice\Invoice {

	protected $zugferd = null;

	public function get_specification() {
		return 'zugferd';
	}

	public function __construct( $profile, $invoice_type = 'invoice' ) {
		parent::__construct( $profile, $invoice_type );

		$this->zugferd = ZugferdDocumentBuilder::CreateNew( $this->map_profile( $this->get_profile() ) );
	}

	/**
	 * @param \Vendidero\StoreaBill\Invoice\Invoice $invoice
	 *
	 * @return string
	 */
	protected function get_invoice_tax_profile( $invoice ) {
		$tax_profile = 'S';

		if ( BusinessInformation::is_small_business() ) {
			$tax_profile = 'O';
		} elseif ( $invoice->is_eu_intra_community_supply() ) {
			$tax_profile = 'K';
		} elseif ( $invoice->is_reverse_charge() ) {
			$tax_profile = 'AE';
		} elseif ( $invoice->is_vat_exempt() ) {
			$tax_profile = 'O';
		} elseif ( $invoice->is_third_country() && 0.0 === $invoice->get_total_tax() ) {
			$tax_profile = 'G';
		}

		return $tax_profile;
	}

	/**
	 * @param $profile
	 *
	 * @see https://docs.peppol.eu/poacc/billing/3.0/codelist/vatex/
	 *
	 * @return null[]|string[]
	 */
	protected function get_tax_exemption_reason_by_profile( $profile ) {
		$exempt_reason      = null;
		$exempt_reason_code = null;

		if ( 'O' === $profile ) {
			$exempt_reason      = 'Not subject to VAT';
			$exempt_reason_code = 'VATEX-EU-IC';
		} elseif ( 'K' === $profile ) {
			$exempt_reason      = 'Intra-Community supply';
			$exempt_reason_code = 'VATEX-EU-IC';
		} elseif ( 'AE' === $profile ) {
			$exempt_reason      = 'Reverse charge';
			$exempt_reason_code = 'VATEX-EU-AE';
		} elseif ( 'G' === $profile ) {
			$exempt_reason      = 'Export outside the EU';
			$exempt_reason_code = 'VATEX-EU-G';
		}

		return array(
			'reason' => $exempt_reason,
			'code'   => $exempt_reason_code,
		);
	}

	/**
	 * @param Item $item
	 *
	 * @see https://service.unece.org/trade/uncefact/vocabulary/rec20/
	 *
	 * @return string
	 */
	protected function get_invoice_item_quantity_code( $item ) {
		$item_code = 'H87'; // piece

		return $item_code;
	}

	public function set_routing_id( $routing_id ) {
		$this->zugferd->setDocumentRoutingId( $routing_id );
	}

	/**
	 * @param \Vendidero\StoreaBill\Invoice\Invoice $invoice
	 * @param string[] $additional_fields
	 *
	 * @return void
	 */
	public function from_invoice( $invoice, $additional_fields = array() ) {
		parent::from_invoice( $invoice, $additional_fields );

		ZugferdSettings::setUnitAmountDecimals( 5 );

		$document_type = 'cancellation' === $this->get_invoice_type() ? '384' : '388';

		$this->zugferd->setDocumentInformation( $invoice->get_formatted_number() ? $invoice->get_formatted_number() : $invoice->get_id(), $document_type, $invoice->get_date_created(), $invoice->get_currency() );
		$this->zugferd->setDocumentSeller( $this->get_seller()->get_legal_name() );

		if ( $this->get_seller()->get_company_id() && $this->get_seller()->get_company_id_scheme() ) {
			$this->zugferd->addDocumentSellerGlobalId( $this->get_seller()->get_company_id(), $this->get_seller()->get_company_id_scheme() );
		}

		if ( $this->get_seller()->get_vat_id() ) {
			$this->zugferd->addDocumentSellerTaxRegistration( 'VA', $this->get_seller()->get_vat_id() );
		}

		if ( $this->get_seller()->get_local_tax_number() ) {
			$this->zugferd->addDocumentSellerTaxRegistration( 'FC', $this->get_seller()->get_local_tax_number() );
		}

		$this->zugferd->setDocumentSellerAddress( $this->get_seller()->get_address_1(), $this->get_seller()->get_address_2(), '', $this->get_seller()->get_postcode(), $this->get_seller()->get_city(), $this->get_seller()->get_country() );
		$this->zugferd->setDocumentSellerContact( $this->get_seller()->get_formatted_full_name() ? $this->get_seller()->get_formatted_full_name() : $this->get_seller()->get_legal_name(), 'Buchhaltung', $this->get_seller()->get_phone(), '', $this->get_seller()->get_email() );
		$this->zugferd->setDocumentSellerCommunication( 'EM', $this->get_seller()->get_email() );

		$this->zugferd->setDocumentBuyer( $this->get_buyer()->get_legal_name(), $this->get_buyer()->get_customer_id() );
		$this->zugferd->setDocumentBuyerAddress( $this->get_buyer()->get_address_1(), $this->get_buyer()->get_address_2(), '', $this->get_buyer()->get_postcode(), $this->get_buyer()->get_city(), $this->get_buyer()->get_country() );
		$this->zugferd->setDocumentBuyerContact( $this->get_buyer()->get_formatted_full_name(), 'Buchhaltung', $this->get_buyer()->get_phone(), '', $this->get_buyer()->get_email() );
		$this->zugferd->setDocumentBuyerCommunication( 'EM', $this->get_buyer()->get_email() );

		if ( $this->get_buyer()->get_company_id() && $this->get_buyer()->get_company_id_scheme() ) {
			$this->zugferd->addDocumentBuyerGlobalId( $this->get_buyer()->get_company_id(), $this->get_buyer()->get_company_id_scheme() );
		}

		if ( $this->get_buyer()->get_vat_id() ) {
			$this->zugferd->addDocumentBuyerTaxRegistration( 'VA', $this->get_buyer()->get_vat_id() );
		}

		if ( $this->get_buyer()->get_local_tax_number() ) {
			$this->zugferd->addDocumentBuyerTaxRegistration( 'FC', $this->get_buyer()->get_local_tax_number() );
		}

		$this->zugferd->setDocumentShipTo( $this->get_ship_to()->get_legal_name(), $this->get_ship_to()->get_customer_id() );
		$this->zugferd->setDocumentShipToAddress( $this->get_ship_to()->get_address_1(), $this->get_ship_to()->get_address_2(), '', $this->get_ship_to()->get_postcode(), $this->get_ship_to()->get_city(), $this->get_ship_to()->get_country() );
		$this->zugferd->setDocumentShipToContact( $this->get_ship_to()->get_formatted_full_name(), '', $this->get_ship_to()->get_phone(), '', $this->get_ship_to()->get_email() );

		if ( $this->get_ship_to()->get_company_id() && $this->get_ship_to()->get_company_id_scheme() ) {
			$this->zugferd->addDocumentShipToGlobalId( $this->get_ship_to()->get_company_id(), $this->get_ship_to()->get_company_id_scheme() );
		}

		if ( $this->get_ship_to()->get_vat_id() ) {
			$this->zugferd->addDocumentShipToTaxRegistration( 'VA', $this->get_ship_to()->get_vat_id() );
		}

		if ( $this->get_ship_to()->get_local_tax_number() ) {
			$this->zugferd->addDocumentShipToTaxRegistration( 'FC', $this->get_ship_to()->get_local_tax_number() );
		}

		$this->zugferd->setDocumentBillingPeriod( $invoice->get_date_of_service(), $invoice->get_date_of_service_end(), '35' );

		if ( 'bacs' === $invoice->get_payment_method_name() ) {
			$payee_iban = Package::get_setting( 'bank_account_iban' );
			$this->zugferd->addDocumentPaymentMean( ZugferdPaymentMeans::UNTDID_4461_58, null, null, null, null, null, $payee_iban );
		} else {
			$this->zugferd->addDocumentPaymentMean( ZugferdPaymentMeans::UNTDID_4461_1 );
		}

		if ( $invoice->is_paid() ) {
			$payment_term = sprintf( _x( 'Invoice paid via %1$s.', 'storeabill-core', 'woocommerce-germanized-pro' ), $invoice->get_payment_method_title() );
		} else {
			$date_due = $invoice->get_date_due() ? $invoice->get_date_due() : $invoice->get_date_created();

			$payment_term = sprintf( _x( 'Pay via %1$s until %2$s.', 'storeabill-core', 'woocommerce-germanized-pro' ), $invoice->get_payment_method_title(), $date_due->date_i18n( get_option( 'date_format' ) ) );
		}

		$this->zugferd->addDocumentPaymentTerm( $payment_term, ! $invoice->is_paid() && $invoice->get_date_due() ? $invoice->get_date_due() : null );

		$item_count     = 0;
		$line_net_total = 0.0;
		$tax_profile    = $this->get_invoice_tax_profile( $invoice );

		foreach ( $invoice->get_items( $invoice->get_item_types_for_totals() ) as $item ) {
			if ( ! is_a( $item, 'Vendidero\StoreaBill\Invoice\TaxableItem' ) ) {
				continue;
			}

			$this->zugferd->addNewPosition( ++$item_count );

			if ( 'product' === $item->get_item_type() ) {
				$this->zugferd->setDocumentPositionProductDetails( $item->get_name(), '', $item->get_sku(), null, '0160', '4012345001235' );

				if ( $item->get_attributes() ) {
					foreach ( $item->get_attributes() as $attribute ) {
						$this->zugferd->addDocumentPositionProductCharacteristic( $attribute->get_formatted_label(), wp_strip_all_tags( $attribute->get_formatted_value() ) );
					}
				}
			} else {
				$this->zugferd->setDocumentPositionProductDetails( $item->get_name() );
			}

			$quantity_code          = $this->get_invoice_item_quantity_code( $item );
			$item_line_subtotal     = sab_add_number_precision( $item->get_subtotal_net(), true, 5 );
			$item_line_total        = sab_add_number_precision( $item->get_total_net(), true, 5 );
			$item_subtotal          = $item_line_subtotal / (int) $item->get_quantity();
			$item_total             = $item_line_total / (int) $item->get_quantity();
			$item_subtotal_quantity = $item_subtotal < 0 ? $item->get_quantity() * -1 : $item->get_quantity();
			$item_total_quantity    = $item_total < 0 ? $item->get_quantity() * -1 : $item->get_quantity();

			if ( sab_remove_number_precision( $item_subtotal, 5 ) !== sab_remove_number_precision( $item_total, 5 ) ) {
				$this->zugferd->setDocumentPositionGrossPrice( abs( sab_remove_number_precision( $item_subtotal, 5 ) ), $item_subtotal_quantity, $quantity_code );
			}

			$this->zugferd->setDocumentPositionNetPrice( abs( sab_remove_number_precision( $item_total, 5 ) ), $item_total_quantity, $quantity_code );
			$this->zugferd->setDocumentPositionQuantity( $item_total_quantity, $quantity_code );

			$line_total_rounded = Numbers::round_to_precision( sab_remove_number_precision( $item_line_total, 5 ), 2 );
			$this->zugferd->setDocumentPositionLineSummation( $line_total_rounded );

			$document_item_tax_profile = $tax_profile;

			if ( $item->get_tax_rates() ) {
				$exempt_data = $this->get_tax_exemption_reason_by_profile( $document_item_tax_profile );

				foreach ( $item->get_tax_rates() as $tax_rate ) {
					$this->zugferd->addDocumentPositionTax( $tax_profile, 'VAT', $tax_rate->get_percent(), null, $exempt_data['reason'], $exempt_data['code'] );
				}
			} else {
				if ( 'S' === $document_item_tax_profile ) {
					$document_item_tax_profile = 'O';
				}

				$exempt_data = $this->get_tax_exemption_reason_by_profile( $document_item_tax_profile );

				$this->zugferd->addDocumentPositionTax( $tax_profile, 'VAT', 0.0, null, $exempt_data['reason'], $exempt_data['code'] );
			}

			$line_net_total += $line_total_rounded;
		}

		$tax_total_rounded = 0.0;
		$tax_net_total     = 0.0;

		foreach ( $invoice->get_tax_totals() as $tax_total ) {
			$exempt_data = $this->get_tax_exemption_reason_by_profile( $tax_profile );

			$this->zugferd->addDocumentTax( $tax_profile, 'VAT', $tax_total->get_total_net(), $tax_total->get_total_tax(), $tax_total->get_tax_rate()->get_percent(), $exempt_data['reason'], $exempt_data['code'] );

			$tax_total_rounded += $tax_total->get_total_tax();
			$tax_net_total     += $tax_total->get_total_net();
		}

		foreach ( $this->get_notes_by_invoice( $invoice ) as $note ) {
			$this->zugferd->addDocumentNote( $note );
		}

		$zero_tax_net_total = Numbers::round_to_precision( $line_net_total - $tax_net_total, 2 );

		if ( 0.0 !== $zero_tax_net_total ) {
			$zero_tax_profile = $tax_profile;

			if ( 'S' === $tax_profile ) {
				$zero_tax_profile = 'O';
			}

			$exempt_data = $this->get_tax_exemption_reason_by_profile( $zero_tax_profile );

			$this->zugferd->addDocumentTax( $tax_profile, 'VAT', $zero_tax_net_total, 0.0, 0.0, $exempt_data['reason'], $exempt_data['code'] );
		}

		$due_payable_amount = $invoice->get_total();
		$prepaid_amount     = 0.0;

		if ( $invoice->is_paid() ) {
			$due_payable_amount = 0.0;
			$prepaid_amount     = $invoice->get_total();
		}

		$this->zugferd->setDocumentSummation( $invoice->get_total(), $due_payable_amount, $line_net_total, null, null, $line_net_total, $tax_total_rounded, null, $prepaid_amount );
	}

	/**
	 * @param $xml
	 *
	 * @return true|\WP_Error
	 */
	public function validate( $xml = '' ) {
		if ( ! $xml ) {
			$xml = $this->get_xml();
		}

		$document  = ZugferdDocumentReader::readAndGuessFromContent( $xml );
		$validator = new ZugferdXsdValidator( $document );

		$result = $validator->validate();

		if ( $result->validationPased() ) {
			return true;
		} else {
			$error = new \WP_Error();

			foreach ( $result->validationErrors() as $error_msg ) {
				$error->add( 'validation-error', $error_msg );
			}

			return $error;
		}
	}

	public function get_xml() {
		return $this->zugferd->getContent();
	}

	/**
	 * @param $profile
	 *
	 * @return int
	 */
	protected function map_profile( $profile ) {
		$zugferd_profile = ZugferdProfiles::PROFILE_EXTENDED;

		switch ( $profile ) {
			case 'basic':
				$zugferd_profile = ZugferdProfiles::PROFILE_BASIC;
				break;
			case 'comfort':
				$zugferd_profile = ZugferdProfiles::PROFILE_EN16931;
				break;
			case 'extended':
				$zugferd_profile = ZugferdProfiles::PROFILE_EXTENDED;
				break;
			case 'xrechnung':
				$zugferd_profile = ZugferdProfiles::PROFILE_XRECHNUNG;
				break;
			case 'xrechnung_2':
				$zugferd_profile = ZugferdProfiles::PROFILE_XRECHNUNG_2;
				break;
			case 'xrechnung_3':
				$zugferd_profile = ZugferdProfiles::PROFILE_XRECHNUNG_3;
				break;
		}

		return $zugferd_profile;
	}

	public function supports_pdf_attachment() {
		return true;
	}

	public function get_pdf( $pdf, $xml = '' ) {
		if ( ! $xml ) {
			$xml = $this->get_xml();
		}

		$merger = new ZugferdDocumentPdfMerger( $xml, $pdf );
		$stream = $merger->generateDocument()->downloadString( 'attached.pdf' );

		return $stream;
	}

	/**
	 * Check if a method is callable by checking the underlying object.
	 * Necessary because is_callable checks will always return true for this object
	 * due to overloading __call.
	 *
	 * @param $method
	 *
	 * @return bool
	 */
	public function is_callable( $method ) {
		if ( method_exists( $this, $method ) ) {
			return true;
		} elseif ( is_callable( array( $this->zugferd, $method ) ) ) {
			return true;
		}

		return false;
	}

	/**
	 * Call child methods if the method does not exist.
	 *
	 * @param $method
	 * @param $args
	 *
	 * @return bool|mixed
	 */
	public function __call( $method, $args ) {
		if ( is_callable( array( $this->zugferd, $method ) ) ) {
			return call_user_func_array( array( $this->zugferd, $method ), $args );
		}

		return false;
	}
}
